from .abstract import *
import random


class Slice(AbstractBHV):
    def __init__(self, b):
        self.b = b

    @classmethod
    def rand(cls) -> Self:
        return cls.random(.5)

    @classmethod
    def random(cls, active: float) -> Self:
        assert 0. <= active <= 1.
        return Slice(random.random() > active)

    def permute(self, permutation_id: int) -> Self:
        raise NotImplementedError()

    def swap_halves(self) -> Self:
        raise NotImplementedError()

    def rehash(self) -> Self:
        raise NotImplementedError()

    def __eq__(self, other: Self) -> bool:
        return Slice(self.b == other.b)

    def __xor__(self, other: Self) -> Self:
        return Slice(self.b ^ other.b)

    def __and__(self, other: Self) -> Self:
        return Slice(self.b & other.b)

    def __or__(self, other: Self) -> Self:
        return Slice(self.b | other.b)

    def __invert__(self) -> Self:
        return Slice(not self.b)

    @classmethod
    def _true_majority(cls, vs: list[Self]) -> Self:
        return Slice(sum(v.b for v in vs) > len(vs)//2)

    def active(self) -> int:
        return int(self.b)
Slice.ONE = Slice(True)
Slice.ZERO = Slice(False)



try:
    from typing import Self
except ImportError:
    Self = 'AbstractBHV'

DIMENSION = 8192

from itertools import groupby
from functools import partial
from dataclasses import fields, is_dataclass
from base64 import _urlsafe_encode_translation
import hashlib
import binascii
import sys

sys.setrecursionlimit(75_000)


def stable_hash(d, try_cache=False) -> bytes:
    """
    Python's `hash` is not stable. That's a problem if you want to use it for persistence and inter-run consistency.
    This one is hand rolled, so don't fully trust it. In fact, if you see anything suspicious, please let me know.
    :param d: Something you want to hash
    :return: The 16-byte md5 hash of the bytes of all (nested) items.
    """
    if try_cache and hasattr(d, "__stable_hash"):
        return d.__stable_hash

    stable_hash_try_cache = partial(stable_hash, try_cache=try_cache)

    t = bytes(type(d).__name__, 'utf-8')
    t += bytes([2])
    if isinstance(d, str):
        t += bytes(d, 'utf-8')
    elif isinstance(d, float):
        t += bytes(d.hex(), 'ascii')
    elif isinstance(d, list) or isinstance(d, tuple):
        t += b''.join(map(stable_hash_try_cache, d))
    elif isinstance(d, set) or isinstance(d, frozenset):
        t += b''.join(sorted(map(stable_hash_try_cache, d)))
    elif isinstance(d, dict):
        for k, v in d.items():
            t += stable_hash_try_cache(k)
            t += bytes([1])
            t += stable_hash_try_cache(v)
    elif is_dataclass(d):
        for f in fields(d):
            t += bytes(f.name, 'utf-8')
            t += bytes([1])
            t += stable_hash_try_cache(getattr(d, f.name))
    else:
        t += bytes(d)
    res = hashlib.md5(t, usedforsecurity=False).digest()
    if try_cache:
        try:
            d.__stable_hash = res
        except AttributeError:
            pass
    return res


def stable_hashcode(d, version=0, try_cache=False) -> str:
    """
    A thin wrapper around `stable_hash` which returns a string and incorporates a version.
    :param d: Something you want to hash
    :param version: Increase this number if you want to unconditionally break all hashcodes
    :return: The 24 character base64 (with - and _) string hash
    """
    h = stable_hash(d, try_cache)
    padded = h.rjust(18, version.to_bytes(1, "little"))
    base64 = binascii.b2a_base64(padded, newline=False)
    url_safe = base64.translate(_urlsafe_encode_translation)
    return url_safe.decode()


def nbs(i, w):
    k = 1
    for _ in range(w):
        yield i ^ k
        k <<= 1


def binw(i, w):
    return bin(i)[2:].rjust(w, '0')


def to_bitmask(s):
    return [x == '1' for x in s]


def bin_bitmask(m):
    return ''.join("01"[x] for x in m)


def bitconfigs(n):
    return [to_bitmask(binw(i, n)) for i in range(2**n)]


def unique_by_id(xs):
    return list(reversed({id(x): x for x in reversed(xs)}.values()))


def format_multiple(xs, start="", sep="", end="", indent="", aindent="", newline_threshold=40):
    ss = list(map(str, xs))
    maxlen = max(map(len, ss))
    if maxlen >= newline_threshold:
        return start + sep.rstrip(" ").join("\n" + aindent + indent + s for s in ss) + "\n" + aindent + end
    else:
        return start + sep.join(s for s in ss) + end


def format_list(xs, **kwargs):
    return format_multiple(xs, start="[", sep=", ", end="]", **kwargs)


class IdSet:
    def __init__(self, iterable=()):
        self.data = {id(x) for x in iterable}

    def __contains__(self, item):
        return id(item) in self.data

    def add(self, item):
        self.data.add(id(item))



# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2016 Mika J. Straka
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Created on Tue Mar 29, 2016

Module:
    poibin - Poisson Binomial Distribution

Author:
    Mika Straka

Description:
    Implementation of the Poisson Binomial distribution for the sum of
    independent and not identically distributed random variables as described
    in the reference [Hong2013]_.

    Implemented method:

        * ``pmf``: probability mass function
        * ``cdf``: cumulative distribution function
        * ``pval``: p-value (1 - cdf)

Usage:
    Be ``p`` a list or  NumPy array of success probabilities for ``n``
    non-identically distributed Bernoulli random variables.

    Import the module and create an instance of the distribution with::

        >>> from poibin import PoiBin
        >>> pb = PoiBin(p)

    Be ``x`` a list or NumPy array of different number of successes.
    To obtain the:

    * probability mass function of x, use::

        >>> pb.pmf(x)

    * cumulative distribution function of x, use::

        >>> pb.cdf(x)

    * p-values of x, use::

        >>> pb.pval(x)

    The functions are applied component-wise and a NumPy array of the same
    length as ``x`` is returned.

References:
.. [Hong2013] Yili Hong, On computing the distribution function for the Poisson
    binomial distribution,
    Computational Statistics & Data Analysis, Volume 59, March 2013,
    Pages 41-51, ISSN 0167-9473,
    http://dx.doi.org/10.1016/j.csda.2012.10.006.
"""

import collections
import numpy as np


class PoiBin(object):
    """Poisson Binomial distribution for random variables.

    This class implements the Poisson Binomial distribution for Bernoulli
    trials with different success probabilities. The distribution describes
    thus a random variable that is the sum of independent and not identically
    distributed single Bernoulli random variables.

    The class offers methods for calculating the probability mass function, the
    cumulative distribution function, and p-values for right-sided testing.
    """

    def __init__(self, probabilities):
        """Initialize the class and calculate the ``pmf`` and ``cdf``.

        :param probabilities: sequence of success probabilities :math:`p_i \\in
            [0, 1] \\forall i \\in [0, N]` for :math:`N` independent but not
            identically distributed Bernoulli random variables
        :type probabilities: numpy.array
        """
        self.success_probabilities = np.array(probabilities)
        self.number_trials = self.success_probabilities.size
        self.check_input_prob()
        self.omega = 2 * np.pi / (self.number_trials + 1)
        self.pmf_list = self.get_pmf_xi()
        self.cdf_list = self.get_cdf(self.pmf_list)

# ------------------------------------------------------------------------------
# Methods for the Poisson Binomial Distribution
# ------------------------------------------------------------------------------

    def pmf(self, number_successes):
        """Calculate the probability mass function ``pmf`` for the input values.

        The ``pmf`` is defined as

        .. math::

            pmf(k) = Pr(X = k), k = 0, 1, ..., n.

        :param number_successes: number of successful trials for which the
            probability mass function is calculated
        :type number_successes: int or list of integers
        """
        self.check_rv_input(number_successes)
        return self.pmf_list[number_successes]

    def cdf(self, number_successes):
        """Calculate the cumulative distribution function for the input values.

        The cumulative distribution function ``cdf`` for a number ``k`` of
        successes is defined as

        .. math::

            cdf(k) = Pr(X \\leq k), k = 0, 1, ..., n.

        :param number_successes: number of successful trials for which the
            cumulative distribution function is calculated
        :type number_successes: int or list of integers
        """
        self.check_rv_input(number_successes)
        return self.cdf_list[number_successes]

    def pval(self, number_successes):
        """Return the p-values corresponding to the input numbers of successes.

        The p-values for right-sided testing are defined as

        .. math::

            pval(k) = Pr(X \\geq k ),  k = 0, 1, ..., n.

        .. note::

            Since :math:`cdf(k) = Pr(X <= k)`, the function returns

            .. math::

                1 - cdf(X < k) & = 1 - cdf(X <= k - 1)
                               & = 1 - cdf(X <= k) + pmf(X = k),

                               k = 0, 1, .., n.

        :param number_successes: number of successful trials for which the
            p-value is calculated
        :type number_successes: int, numpy.array, or list of integers
        """
        self.check_rv_input(number_successes)
        i = 0
        try:
            isinstance(number_successes, collections.Iterable)
            pvalues = np.array(number_successes, dtype='float')
            # if input is iterable (list, numpy.array):
            for k in number_successes:
                pvalues[i] = 1. - self.cdf(k) + self.pmf(k)
                i += 1
            return pvalues
        except TypeError:
            # if input is an integer:
            if number_successes == 0:
                return 1
            else:
                return 1 - self.cdf(number_successes - 1)

# ------------------------------------------------------------------------------
# Methods to obtain pmf and cdf
# ------------------------------------------------------------------------------

    def get_cdf(self, event_probabilities):
        """Return the values of the cumulative density function.

        Return a list which contains all the values of the cumulative
        density function for :math:`i = 0, 1, ..., n`.

        :param event_probabilities: array of single event probabilities
        :type event_probabilities: numpy.array
        """
        cdf = np.empty(self.number_trials + 1)
        cdf[0] = event_probabilities[0]
        for i in range(1, self.number_trials + 1):
            cdf[i] = cdf[i - 1] + event_probabilities[i]
        return cdf

    def get_pmf_xi(self):
        """Return the values of the variable ``xi``.

        The components ``xi`` make up the probability mass function, i.e.
        :math:`\\xi(k) = pmf(k) = Pr(X = k)`.
        """
        chi = np.empty(self.number_trials + 1, dtype=complex)
        chi[0] = 1
        half_number_trials = int(
            self.number_trials / 2 + self.number_trials % 2)
        # set first half of chis:
        chi[1:half_number_trials + 1] = self.get_chi(
            np.arange(1, half_number_trials + 1))
        # set second half of chis:
        chi[half_number_trials + 1:self.number_trials + 1] = np.conjugate(
            chi[1:self.number_trials - half_number_trials + 1] [::-1])
        chi /= self.number_trials + 1
        xi = np.fft.fft(chi)
        if self.check_xi_are_real(xi):
            xi = xi.real
        else:
            raise TypeError("pmf / xi values have to be real.")
        xi += np.finfo(type(xi[0])).eps
        return xi

    def get_chi(self, idx_array):
        """Return the values of ``chi`` for the specified indices.

        :param idx_array: array of indices for which the ``chi`` values should
            be calculated
        :type idx_array: numpy.array
        """
        # get_z:
        exp_value = np.exp(self.omega * idx_array * 1j)
        xy = 1 - self.success_probabilities + \
            self.success_probabilities * exp_value[:, np.newaxis]
        # sum over the principal values of the arguments of z:
        argz_sum = np.arctan2(xy.imag, xy.real).sum(axis=1)
        # get d value:
        exparg = np.log(np.abs(xy)).sum(axis=1)
        d_value = np.exp(exparg)
        # get chi values:
        chi = d_value * np.exp(argz_sum * 1j)
        return chi

# ------------------------------------------------------------------------------
# Auxiliary functions
# ------------------------------------------------------------------------------

    def check_rv_input(self, number_successes):
        """Assert that the input values ``number_successes`` are OK.

        The input values ``number_successes`` for the random variable have to be
        integers, greater or equal to 0, and smaller or equal to the total
        number of trials ``self.number_trials``.

        :param number_successes: number of successful trials
        :type number_successes: int or list of integers """
        try:
            for k in number_successes:
                assert (type(k) == int or type(k) == np.int64), \
                        "Values in input list must be integers"
                assert k >= 0, 'Values in input list cannot be negative.'
                assert k <= self.number_trials, \
                    'Values in input list must be smaller or equal to the ' \
                    'number of input probabilities "n"'
        except TypeError:
            assert (type(number_successes) == int or \
                type(number_successes) == np.int64), \
                'Input value must be an integer.'
            assert number_successes >= 0, "Input value cannot be negative."
            assert number_successes <= self.number_trials, \
                'Input value cannot be greater than ' + str(self.number_trials)
        return True

    @staticmethod
    def check_xi_are_real(xi_values):
        """Check whether all the ``xi``s have imaginary part equal to 0.

        The probabilities :math:`\\xi(k) = pmf(k) = Pr(X = k)` have to be
        positive and must have imaginary part equal to zero.

        :param xi_values: single event probabilities
        :type xi_values: complex
        """
        return np.all(xi_values.imag <= np.finfo(float).eps)

    def check_input_prob(self):
        """Check that all the input probabilities are in the interval [0, 1]."""
        if self.success_probabilities.shape != (self.number_trials,):
            raise ValueError(
                "Input must be an one-dimensional array or a list.")
        if not np.all(self.success_probabilities >= 0):
            raise ValueError("Input probabilities have to be non negative.")
        if not np.all(self.success_probabilities <= 1):
            raise ValueError("Input probabilities have to be smaller than 1.")



from .abstract import *
import numpy as np
from sys import byteorder, version_info


class NumPyBoolPermutation(MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}
    rng = np.random.default_rng()

    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def random(cls) -> 'NumPyBoolPermutation':
        return NumPyBoolPermutation(cls.rng.permutation(DIMENSION))

    def __mul__(self, other: 'NumPyBoolPermutation') -> 'NumPyBoolPermutation':
        return NumPyBoolPermutation(self.data[other.data])

    def __invert__(self) -> 'NumPyBoolPermutation':
        inv_permutation = np.empty_like(self.data)
        inv_permutation[self.data] = np.arange(DIMENSION)
        return NumPyBoolPermutation(inv_permutation)

    def __call__(self, hv: 'NumPyBoolBHV') -> 'NumPyBoolBHV':
        return hv.permute_bits(self)

NumPyBoolPermutation.IDENTITY = NumPyBoolPermutation(np.arange(DIMENSION))


class NumPyBoolBHV(AbstractBHV):
    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def rand(cls) -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.random.randint(0, high=2, size=DIMENSION, dtype=np.bool_))

    @classmethod
    def random(cls, active: float) -> 'NumPyBoolBHV':
        assert 0. <= active <= 1.
        return NumPyBoolBHV(np.random.binomial(1, active, DIMENSION))

    def select(self, when1: 'NumPyBoolBHV', when0: 'NumPyBoolBHV') -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.where(self.data, when1.data, when0.data))

    def swap_halves(self) -> 'NumPyBoolBHV':
        return self.roll_bits(DIMENSION//2)

    def rehash(self) -> 'NumPyBoolBHV':
        return self.pack8().rehash().unpack()

    def roll_bits(self, n: int) -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.roll(self.data, n))

    def permute_bits(self, permutation: 'NumPyBoolPermutation') -> 'NumPyBoolBHV':
        return NumPyBoolBHV(self.data[permutation.data])

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> 'NumPyBoolBHV':
        return self.permute_bits(NumPyBoolPermutation.get(permutation_id))

    @classmethod
    def majority(cls, vs: list['NumPyBoolBHV']) -> 'NumPyBoolBHV':
        data = [v.data for v in vs]
        extra = [cls.rand().data] if len(vs) % 2 == 0 else []

        tensor = np.stack(data + extra)
        counts = tensor.sum(axis=-2, dtype=np.uint8 if len(vs) < 256 else np.uint32)

        threshold = (len(vs) + len(extra))//2

        return NumPyBoolBHV(np.greater(counts, threshold))

    def __eq__(self, other: 'NumPyBoolBHV') -> bool:
        return np.array_equal(self.data, other.data)

    def __xor__(self, other: 'NumPyBoolBHV') -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.bitwise_xor(self.data, other.data))

    def __and__(self, other: 'NumPyBoolBHV') -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.bitwise_and(self.data, other.data))

    def __or__(self, other: 'NumPyBoolBHV') -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.bitwise_or(self.data, other.data))

    def __invert__(self) -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.bitwise_not(self.data))

    def active(self) -> int:
        return int(np.sum(self.data))

    def pack8(self) -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.packbits(self.data))

    def pack64(self) -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(np.packbits(self.data).view(dtype=np.uint64))

    def to_bytes(self):
        return self.pack8().to_bytes()

    @classmethod
    def from_bytes(cls, bs):
        return NumPyPacked8BHV.from_bytes(bs).unpack()

    def bits(self):
        return iter(self.data.astype(np.uint8))

    def bitstring(self) -> str:
        b = ord('0')
        return (self.data.astype(np.uint8) + b).tobytes().decode('ascii')

    @classmethod
    def from_bitstring(cls, s: str) -> str:
        b = ord('0')
        return cls((np.frombuffer(bytes(s, 'ascii'), dtype=np.uint8) - b).astype(np.bool_))

NumPyBoolBHV.ZERO = NumPyBoolBHV(np.zeros(DIMENSION, dtype=np.bool_))
NumPyBoolBHV.ONE = NumPyBoolBHV(np.ones(DIMENSION, dtype=np.bool_))
NumPyBoolBHV._FEISTAL_SUBKEYS = NumPyBoolBHV.nrand2(NumPyBoolBHV._FEISTAL_ROUNDS, 4)
_halfb = np.zeros(DIMENSION, dtype=np.bool_)
_halfb[:DIMENSION//2] = np.bool_(True)
NumPyBoolBHV.HALF = NumPyBoolBHV(_halfb)


class NumPyBytePermutation(MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}
    rng = np.random.default_rng()

    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def random(cls) -> 'NumPyBytePermutation':
        return NumPyBytePermutation(cls.rng.permutation(DIMENSION//8))

    def __mul__(self, other: 'NumPyBytePermutation') -> 'NumPyBytePermutation':
        return NumPyBytePermutation(self.data[other.data])

    def __invert__(self) -> 'NumPyBytePermutation':
        inv_permutation = np.empty_like(self.data)
        inv_permutation[self.data] = np.arange(DIMENSION//8)
        return NumPyBytePermutation(inv_permutation)

    def __call__(self, hv: 'NumPyPacked8BHV') -> 'NumPyPacked8BHV':
        return hv.permute_bytes(self)

NumPyBytePermutation.IDENTITY = NumPyBytePermutation(np.arange(DIMENSION//8))


class NumPyPacked8BHV(AbstractBHV):
    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def rand(cls) -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.random.randint(0, 255, DIMENSION//8, dtype=np.uint8))

    @classmethod
    def random(cls, active: float) -> 'NumPyPacked8BHV':
        assert 0. <= active <= 1.
        return NumPyBoolBHV.random(active).pack8()

    def roll_bytes(self, n: int) -> 'NumPyPacked8BHV':
        assert abs(n) < DIMENSION//8, "only supports DIMENSION/8 rolls"
        return NumPyPacked8BHV(np.roll(self.data, n))

    def swap_halves(self) -> 'NumPyPacked8BHV':
        return self.roll_bytes(DIMENSION//16)

    def permute_bytes(self, permutation: 'NumPyBytePermutation') -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(self.data[permutation.data])

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> 'NumPyPacked8BHV':
        return self.permute_bytes(NumPyBytePermutation.get(permutation_id))

    def rehash(self) -> 'NumPyPacked8BHV':
        byte_data = self.to_bytes()
        rehashed_byte_data = hashlib.shake_256(byte_data).digest(DIMENSION//8)
        return NumPyPacked8BHV.from_bytes(rehashed_byte_data)

    def __eq__(self, other: 'NumPyPacked8BHV') -> bool:
        return np.array_equal(self.data, other.data)

    def __xor__(self, other: 'NumPyPacked8BHV') -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.bitwise_xor(self.data, other.data))

    def __and__(self, other: 'NumPyPacked8BHV') -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.bitwise_and(self.data, other.data))

    def __or__(self, other: 'NumPyPacked8BHV') -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.bitwise_or(self.data, other.data))

    def __invert__(self) -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(np.bitwise_not(self.data))

    if version_info[2] >= 10:
        def active(self) -> int:
            return int.from_bytes(self.data.tobytes(), byteorder).bit_count()
    else:
        lookup = np.array([bin(i).count("1") for i in range(256)])

        def active(self) -> int:
            return self.lookup[self.data].sum()

    def unpack(self) -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.unpackbits(self.data))

    def repack64(self) -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(self.data.view(dtype=np.uint64))

    def to_bytes(self):
        return self.data.tobytes()

    @classmethod
    def from_bytes(cls, bs):
        return cls(np.frombuffer(bs, dtype=np.uint8))


NumPyPacked8BHV.ZERO = NumPyPacked8BHV(np.zeros(DIMENSION//8, dtype=np.uint8))
NumPyPacked8BHV.ONE = NumPyPacked8BHV(np.full(DIMENSION//8, fill_value=255, dtype=np.uint8))
NumPyPacked8BHV._FEISTAL_SUBKEYS = NumPyPacked8BHV.nrand2(NumPyPacked8BHV._FEISTAL_ROUNDS, 4)
_half8 = np.zeros(DIMENSION//8, dtype=np.uint8)
_half8[:DIMENSION//16] = np.uint8(255)
NumPyPacked8BHV.HALF = NumPyPacked8BHV(_half8)


class NumPyWordPermutation(MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}
    rng = np.random.default_rng()

    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def random(cls) -> 'NumPyWordPermutation':
        return NumPyWordPermutation(cls.rng.permutation(DIMENSION//64))

    def __mul__(self, other: 'NumPyWordPermutation') -> 'NumPyWordPermutation':
        return NumPyWordPermutation(self.data[other.data])

    def __invert__(self) -> 'NumPyWordPermutation':
        inv_permutation = np.empty_like(self.data)
        inv_permutation[self.data] = np.arange(DIMENSION//64)
        return NumPyWordPermutation(inv_permutation)

    def __call__(self, hv: 'NumPyPacked64BHV') -> 'NumPyPacked64BHV':
        return hv.permute_words(self)

NumPyWordPermutation.IDENTITY = NumPyWordPermutation(np.arange(DIMENSION//64))


class NumPyPacked64BHV(AbstractBHV):
    rng = np.random.SFC64()

    def __init__(self, array: np.ndarray):
        self.data: np.ndarray = array

    @classmethod
    def rand(cls) -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(cls.rng.random_raw(DIMENSION//64))

    @classmethod
    def random(cls, active: float) -> 'NumPyPacked64BHV':
        assert 0. <= active <= 1.
        return NumPyBoolBHV.random(active).pack64()

    @classmethod
    def majority(cls, vs: list['NumPyPacked64BHV']) -> 'NumPyPacked64BHV':
        if len(vs) <= 9:
            return cls._majority_via_custom(vs)
        else:
            return cls._majority_via_unpacked(vs)

    @classmethod
    def _majority_via_unpacked(cls, vs: list['NumPyPacked64BHV']) -> 'NumPyPacked64BHV':
        return NumPyBoolBHV.majority([v.unpack() for v in vs]).pack64()

    def swap_halves(self) -> 'NumPyPacked64BHV':
        return self.roll_words(DIMENSION//128)

    def rehash(self) -> 'NumPyPacked64BHV':
        return self.repack8().rehash().repack64()

    def roll_words(self, n: int) -> 'NumPyPacked64BHV':
        assert abs(n) < DIMENSION//64, "only supports DIMENSION/64 rolls"
        return NumPyPacked64BHV(np.roll(self.data, n))

    def roll_word_bits(self, n: int) -> 'NumPyPacked64BHV':
        assert abs(n) < 64, "only supports 64 rolls"
        # https://github.com/numba/numba/issues/6381
        if n == 0:
            return NumPyPacked64BHV(self.data)
        elif n > 0:
            return NumPyPacked64BHV(np.bitwise_or(np.right_shift(self.data, np.uint64(n)), np.left_shift(self.data, np.uint64(64 - n))))
        else:
            return NumPyPacked64BHV(np.bitwise_or(np.left_shift(self.data, np.uint64(n)), np.right_shift(self.data, np.uint64(64 - n))))

    # roll_words and roll_word_bits could be combined for more options allowing positive and negative combinations
    # ((1 2 3 4) (a b c d) (α β γ δ))
    # rolled by 1, -2 for example results in
    # ((γ δ α β) (3 4 1 2) (c d a b))

    def permute_words(self, permutation: 'NumPyWordPermutation') -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(self.data[permutation.data])

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> 'NumPyPacked64BHV':
        return self.permute_words(NumPyWordPermutation.get(permutation_id))

    def __eq__(self, other: 'NumPyPacked64BHV') -> bool:
        return np.array_equal(self.data, other.data)

    def __xor__(self, other: 'NumPyPacked64BHV') -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(np.bitwise_xor(self.data, other.data))

    def __and__(self, other: 'NumPyPacked64BHV') -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(np.bitwise_and(self.data, other.data))

    def __or__(self, other: 'NumPyPacked64BHV') -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(np.bitwise_or(self.data, other.data))

    def __invert__(self) -> 'NumPyPacked64BHV':
        return NumPyPacked64BHV(np.bitwise_not(self.data))

    if version_info[2] >= 10:
        def active(self) -> int:
            return int.from_bytes(self.data.tobytes(), byteorder).bit_count()
    else:
        def active(self) -> int:
            return self.repack8().active()

    def unpack(self) -> 'NumPyBoolBHV':
        return NumPyBoolBHV(np.unpackbits(self.data.view(np.uint8)))

    def repack8(self) -> 'NumPyPacked8BHV':
        return NumPyPacked8BHV(self.data.view(dtype=np.uint8))

    def to_bytes(self):
        return self.repack8().to_bytes()

    @classmethod
    def from_bytes(cls, bs):
        return NumPyPacked8BHV.from_bytes(bs).repack64()

NumPyPacked64BHV.ZERO = NumPyPacked64BHV(np.zeros(DIMENSION//64, dtype=np.uint64))
NumPyPacked64BHV.ONE = NumPyPacked64BHV(np.full(DIMENSION//64, fill_value=-1, dtype=np.uint64))
NumPyPacked64BHV._FEISTAL_SUBKEYS = NumPyPacked64BHV.nrand2(NumPyPacked64BHV._FEISTAL_ROUNDS, 4)
_half64 = np.zeros(DIMENSION//64, dtype=np.uint64)
_half64[:DIMENSION//128] = np.uint64(-1)
NumPyPacked64BHV.HALF = NumPyPacked64BHV(_half64)



from typing import Generic, TypeVar, Type
from copy import deepcopy
from .abstract import *

HV = TypeVar("HV", bound=BooleanAlgBHV, covariant=True)


class LinearBHVModel(Generic[HV]):
    def __init__(self, bhv: Type[HV]):
        self.bhv = bhv
        self.used = IdSet()
    
    def spawn(self, one: 'HV.ONE', zero: 'HV.ZERO') -> (HV, HV):
        assert one not in self.used and zero not in self.used
        self.used.add(one); self.used.add(zero)
        # I O  ->  P N
        # 1 0  ->  1 0  |  0 1
        r = self.bhv.select_rand(one, zero)
        return (r, ~r)

    def xor(self, x: HV, y: HV) -> (HV, HV, HV):
        assert x not in self.used and y not in self.used
        self.used.add(x); self.used.add(y)
        # L R  ->  X L R
        # 0 0  ->  0 0 0
        # 0 1  ->  1 0 0
        # 1 0  ->  1 0 0
        # 1 1  ->  0 1 1
        b = x & y
        return (x ^ y, b, deepcopy(b))

    def thresholds3(self, x: 'HV', y: 'HV', z: 'HV') -> ('HV', 'HV', 'HV'):
        assert x not in self.used and y not in self.used and z not in self.used
        self.used.add(x); self.used.add(y); self.used.add(z)
        # x y z  ->  + M -
        # x y z  ->  + M -
        # 0 0 0  ->  0 0 0

        # 0 0 1  ->  0 0 1
        # 0 1 0  ->  0 0 1
        # 1 0 0  ->  0 0 1

        # 0 1 1  ->  0 1 1
        # 1 0 1  ->  0 1 1
        # 1 1 0  ->  0 1 1

        # 1 1 1  ->  1 1 1
        return (x & y & z, self.bhv.majority([x, y, z]), x | y | z)

    def and_or(self, x: 'HV', y: 'HV') -> ('HV', 'HV'):
        assert x not in self.used and y not in self.used
        self.used.add(x); self.used.add(y)
        # x y  ->  & |
        # 0 0  ->  0 0
        # 0 1  ->  0 1
        # 1 0  ->  0 1
        # 1 1  ->  1 1
        return (x & y, x | y)

    def switch(self, cond: 'HV', left: 'HV', right: 'HV') -> ('HV', 'HV', 'HV'):
        assert cond not in self.used and left not in self.used and right not in self.used
        self.used.add(cond); self.used.add(left); self.used.add(right)

        # C L R  ->  C P N
        # 0 0 0  ->  0 0 0

        # 0 0 1  ->  0 0 1
        # 0 1 0  ->  0 1 0
        # 0 1 1  ->  0 1 1

        # 1 0 1  ->  0 1 0
        # 1 1 0  ->  0 0 1
        # 1 1 1  ->  0 1 1

        # 1 1 1  ->  1 1 1

        pos = self.bhv.select(cond, left, right)
        neg = self.bhv.select(cond, right, left)
        return (deepcopy(cond), pos, neg)

    def invert(self, x: 'HV', one: 'HV.ONE', zero: 'HV.ZERO') -> ('HV', 'HV', 'HV'):
        # P I O  ->  N A B
        # 0 1 0  ->  1 0 0
        # 1 1 0  ->  0 1 1
        (self_1, self_2, inverted) = self.switch(x, one, zero)
        return (inverted, self_1, self_2)

    def permute(self, x: 'HV', permutation_id: int) -> 'HV':
        assert x not in self.used
        self.used.add(x)
        return self.bhv.permute(permutation_id)


class Tank:
    def __init__(self, capacity: float, fill=None, check=True, record=False):
        self.capacity = capacity
        self.level = capacity if fill is None else fill
        self.check = check
        self.record = record
        self.updates = []

    def update(self, amount: float):
        if self.check:
            if amount < 0:
                assert self.level >= amount, "not enough to drain"
            else:
                assert amount <= self.capacity, "too much to fill"
        if self.record:
            self.updates.append(amount)
        self.level += amount

    def historical_levels(self):
        assert self.record, "history not recorded, please pass record=True"
        level = self.level
        record = [level]
        for update in reversed(self.updates):
            level -= update
            record.append(level)
        record.reverse()
        return record


class AdiabaticBHVModel:
    def __init__(self, tank: Tank, bhv: Type[HV]):
        self.tank = tank
        self.bhv = bhv
        self.used = IdSet()

    def get_one(self) -> 'HV.ONE':
        self.tank.update(-DIMENSION)
        return deepcopy(self.bhv.ONE)

    def get_zero(self) -> 'HV.ZERO':
        self.tank.update(0)
        return deepcopy(self.bhv.ZERO)

    def spawn(self) -> ('HV', 'HV'):
        # Note: maybe there's a more efficient way to generate random vectors in this context
        # I O  ->  P N
        # 1 0  ->  1 0  |  0 1
        r = self.bhv.rand()
        self.tank.update(-DIMENSION)
        return (r, ~r)

    def xor(self, x: 'HV', y: 'HV') -> 'HV':
        assert x not in self.used and y not in self.used
        self.used.add(x); self.used.add(y)
        # L R  ->  X
        # 0 0  ->  0
        # 0 1  ->  1
        # 1 0  ->  1
        # 1 1  ->  0
        b = (x & y).active()
        self.tank.update(2*b)
        return x ^ y

    def majority3(self, x: 'HV', y: 'HV', z: 'HV') -> 'HV':
        assert x not in self.used and y not in self.used and z not in self.used
        self.used.add(x); self.used.add(y); self.used.add(z)
        # x y z  ->  M
        # 0 0 0  ->  0

        # 0 0 1  ->  0
        # 0 1 0  ->  0
        # 1 0 0  ->  0

        # 0 1 1  ->  1
        # 1 0 1  ->  1
        # 1 1 0  ->  1

        # 1 1 1  ->  1
        self.tank.update((x & y & z).active() + (x | y | z).active())
        return self.bhv.majority([x, y, z])

    def and_or(self, x: 'HV', y: 'HV') -> ('HV', 'HV'):
        assert x not in self.used and y not in self.used
        self.used.add(x); self.used.add(y)
        # x y  ->  & |
        # 0 0  ->  0 0
        # 0 1  ->  0 1
        # 1 0  ->  0 1
        # 1 1  ->  1 1
        self.tank.update(0)
        return (x & y, x | y)

    def switch(self, cond: 'HV', left: 'HV', right: 'HV') -> ('HV', 'HV', 'HV'):
        assert cond not in self.used and left not in self.used and right not in self.used
        self.used.add(cond); self.used.add(left); self.used.add(right)
        # C L R  ->  C P N
        # 0 0 0  ->  0 0 0

        # 0 0 1  ->  0 0 1
        # 0 1 0  ->  0 1 0
        # 0 1 1  ->  0 1 1

        # 1 0 1  ->  0 1 0
        # 1 1 0  ->  0 0 1
        # 1 1 1  ->  0 1 1

        # 1 1 1  ->  1 1 1

        pos = cond.select(left, right)
        neg = cond.select(right, left)
        self.tank.update(0)
        return (deepcopy(cond), pos, neg)

    def invert(self, x: 'HV') -> 'HV':
        assert x not in self.used
        self.used.add(x)
        # P I O  ->  N A B
        # 0 1 0  ->  1 0 0
        # 1 1 0  ->  0 1 1
        self.tank.update(2*(x.active() - DIMENSION//2))
        return ~x

    def permute(self, permutation_id: int) -> 'HV':
        self.tank.update(0)
        return self.bhv.permute(permutation_id)




from .symbolic import Var, Symbolic
from typing import Optional

Knowledge = dict[str, Symbolic]


# Unify t1 with t2 using/improving the passed knowledge.
def unify(t1: Symbolic, t2: Symbolic, knowledge: Optional[Knowledge] = None) -> Optional[Knowledge]:
    if knowledge is None:
        knowledge = {}

    if isinstance(t1, Var) and isinstance(t2, Var) and t1.name == t2.name:
        return knowledge  # does global variable equality
    elif isinstance(t2, Var):
        return unify_variable(t2.name, t1, knowledge)
    elif isinstance(t1, Var):
        return unify_variable(t1.name, t2, knowledge)
    else:
        if t1.constant() != t2.constant():
            return None
        else:
            cs1 = t1.children()
            cs2 = t2.children()
            if len(cs1) != len(cs2):
                return None
            else:
                k = knowledge

                for l, r in zip(cs1, cs2):
                    k = unify(l, r, k)
                    if k is None:
                        return None
                return k


def unify_variable(v: str, t: Symbolic, knowledge: Knowledge) -> Optional[Knowledge]:
    if v in knowledge:
        return unify(knowledge[v], t, knowledge)
    elif isinstance(t, Var) and t.name in knowledge:
        return unify(Var(v), knowledge[t.name], knowledge)
    elif occurs_check(v, t, knowledge):
        return None
    else:
        return {**knowledge, v: t}


def occurs_check(v: str, t, knowledge: Knowledge):
    if isinstance(t, Var) and t.name == v:
        return True
    elif isinstance(t, Var) and t.name in knowledge:
        return occurs_check(v, knowledge[t.name], knowledge)
    else:
        return any(occurs_check(v, arg, knowledge) for arg in t.children())



from .abstract import AbstractBHV
from typing import TypeVar, Generic, Iterable, Iterator, Optional, Mapping
from math import comb


K = TypeVar("K")


class Store(Generic[K]):
    @classmethod
    def auto_associative(cls, vs: Iterable[AbstractBHV]) -> 'Store[int]':
        return cls(dict(enumerate(vs)))

    def __init__(self, hvs: Mapping[K, AbstractBHV]):
        self.hvs = hvs

    def related(self, v: AbstractBHV, threshold=6) -> Iterator[K]:
        raise NotImplementedError()

    def closest(self, v: AbstractBHV) -> Optional[K]:
        raise NotImplementedError()


class StoreList(Store):
    def related(self, v: AbstractBHV, threshold=6) -> Iterator[K]:
        for k, v_ in self.hvs.items():
            if v_.std_apart(v) <= threshold:
                yield k

    def closest(self, v: AbstractBHV) -> Optional[K]:
        closest = None
        shortest_distance = float('inf')

        for k, v_ in self.hvs.items():
            d = v_.std_apart(v)
            if d < shortest_distance:
                closest = k
                shortest_distance = d

        return closest


class StoreRejectList(Store):
    def __init__(self, hvs: dict[K, AbstractBHV]):
        super().__init__(hvs)
        vs = list(hvs.values())
        representative = vs[0]
        self.bundle = representative.majority(vs)
        self.bundle_size = len(vs)
        self.included_std = AbstractBHV.frac_to_std(AbstractBHV.maj_ber(self.bundle_size))
    # expected number of bits flipped compared to the majority of N random hvs
    # E[bit_error_rate(v, MAJ(v, v_0, ..v_n))]
    # E[v_q != MAJ(v, v_0, ..v_n)_q] WLOG
    # E[1_q != MAJ(1, v_0, ..v_n)_q]
    # E[1 != MAJ(1, v_0, ..v_n)_q]
    # PoiBin(1, af(v_0), ..af(v_n)).cdf(n//2)
    # further assuming af(v_0) == af(v_k) == 0.5,
    # for n=1,3,5,...
    # E[v_q != MAJ(v, v_0, ..v_n)_q]
    # E[MAJ(v, v_0, ..v_n)_q]
    # E[1 | SUM(v_k) > n/2
    #   0 | SUM(v_k) < n/2    # doesn't contribute
    #   v | SUM(v_k) = n/2]   # using af(v) == 0.5
    # in code
    # sum(.5 if sum(bc) == n//2 else (sum(bc) > n//2) for bc in bitconfigs(n))
    # E[SUM(v_k) > n/2] + E[V]
    # (comb(n - 1, (n - 1)//2))/2**n
    # n! / ((n/2)! * (n - n/2)!)

    def related_reject(self, v: AbstractBHV, threshold=6, reject_safety=3) -> Iterator[K]:
        if self.bundle.unrelated(v, self.included_std - reject_safety):
            return
        for k, v_ in self.hvs.items():
            if v_.std_apart(v) <= threshold:
                yield k

    def related(self, v: AbstractBHV, threshold=6) -> Iterator[K]:
        for k, v_ in self.hvs.items():
            if v_.std_apart(v) <= threshold:
                yield k

    def find_closest(self, v: AbstractBHV) -> Optional[K]:
        closest = None
        shortest_distance = float('inf')

        for k, v_ in self.hvs.items():
            d = v_.std_apart(v)
            if d < shortest_distance:
                closest = k
                shortest_distance = d

        return closest


class StoreChunks(Store):
    def __init__(self, hvs: dict[K, AbstractBHV], chunk_size=50):
        super().__init__(hvs)
        ks = list(hvs.keys())
        vs = list(hvs.values())
        representative = vs[0]
        self.chunks = [(representative.majority(vs[i*chunk_size:(i+1)*chunk_size]), dict(zip(ks[i*chunk_size:(i+1)*chunk_size], vs[i*chunk_size:(i+1)*chunk_size])))
                       for i in range(len(vs)//chunk_size+1)]
        self.total_size = len(vs)
        self.chunk_size = chunk_size
        self.included_std = AbstractBHV.frac_to_std(AbstractBHV.maj_ber(chunk_size))

    def related(self, v: AbstractBHV, threshold=6) -> Iterator[K]:
        for maj, chunk in self.chunks:
            if not maj.unrelated(v, self.included_std - 3):
                for k, v_ in chunk.items():
                    if v_.related(v, threshold):
                        yield k

    def find_closest(self, v: AbstractBHV) -> Optional[K]:
        closest = None
        shortest_distance = float('inf')

        for k, v_ in self.hvs.items():
            d = v_.std_apart(v)
            if d < shortest_distance:
                closest = k
                shortest_distance = d

        return closest



from .abstract import AbstractBHV, DIMENSION


class DistanceGraph:
    @classmethod
    def from_scope(cls, local_dict):
        return cls(*zip(*[(v, k) for k, v in local_dict.items() if isinstance(v, AbstractBHV)]))

    def __init__(self, hvs: 'list[AbstractBHV]', labels: 'list[str]'):
        self.hvs = hvs
        self.labels = labels
        self.adj = [[round(min(v.std_apart(w, invert=True), v.std_apart(w))) if not v.unrelated(w) else None
                     for v in self.hvs]
                    for w in self.hvs]

    def graphviz(self):
        for i, (r, l) in enumerate(zip(self.adj, self.labels)):
            print(f"{i} [label=\"{l}\"];")
            for j, d in enumerate(r):
                if i < j and d is not None:
                    print(f"{i} -- {j} [label=\"{d}\"];")


class Image:
    @classmethod
    def load_pbm(cls, file: 'IO[Any]', bhv: 'AbstractBHV', binary=False):
        if binary:
            header = file.readline().rstrip()
            assert header == b"P4"
            dimension, n = map(int, file.readline().rstrip().split(b" "))
            assert dimension == DIMENSION
            hvs = [bhv.from_bytes(file.read(DIMENSION//8)) for _ in range(n)]
            return cls(hvs)
        else:
            header = file.readline().rstrip()
            assert header == "P1"
            dimension, n = map(int, file.readline().rstrip().split(" "))
            assert dimension == DIMENSION
            hvs = [bhv.from_bitstring(file.readline().rstrip()) for _ in range(n)]
            return cls(hvs)

    def __init__(self, hvs: 'list[AbstractBHV]'):
        self.hvs = hvs

    def pbm(self, file: 'IO[Any]', binary=False):
        if binary:
            file.write(b"P4\n" + bytes(str(DIMENSION), "ascii") + b" " + bytes(str(len(self.hvs)), "ascii") + b"\n")
            for hv in self.hvs:
                file.write(hv.to_bytes())
        else:
            file.write(f"P1\n{DIMENSION} {len(self.hvs)}\n")
            for hv in self.hvs:
                file.write(hv.bitstring())
                file.write('\n')



from setuptools import setup, find_packages, Extension

VERSION = '0.6.10a'
DESCRIPTION = 'Boolean Hypervectors'
LONG_DESCRIPTION = 'Boolean Hypervectors with various operators for experiments in hyperdimensional computing (HDC).'

native = Extension("bhv.cnative",
                   sources=['bhv/cnative/bindings.cpp',
                            'bhv/cnative/TurboSHAKEopt/TurboSHAKE.cpp',
                            'bhv/cnative/TurboSHAKEopt/KeccakSponge.cpp',
                            'bhv/cnative/TurboSHAKEopt/KeccakP-1600-opt64.cpp',
                            ],
                   include_dirs=['bhv/cnative', 'bhv/cnative/TurboSHAKEopt'],
                   extra_compile_args=['-std=c++2a', '-O3', '-march=native', '-Wall'],
                   language='c++',
                   optional=True)

setup(
    name="bhv",
    version=VERSION,
    author="Adam Vandervorst",
    author_email="contact@adamv.be",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="https://github.com/Adam-Vandervorst/PyBHV",
    packages=find_packages(),
    install_requires=[],
    extras_require={
        "pytorch": ["torch>=2.0.0"],
        "np": ["numpy>=1.24.2"],
    },
    keywords='ai binary hypervector hdc bsc',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',

        'License :: Free for non-commercial use',

        'Environment :: GPU :: NVIDIA CUDA',

        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',

        'Typing :: Typed',
    ],
    ext_modules=[native]
)



from dataclasses import dataclass, field, fields
from .abstract import *
from .shared import stable_hashcode, bitconfigs, unique_by_id, format_multiple, format_list
from .slice import Slice


class Symbolic:
    name = None

    def named(self, name):
        if name is not None: self.name = name
        return self

    def nodename(self, **kwargs):
        return self.name or f"{type(self).__name__.upper()}"

    def nodeid(self, structural=False, **kwargs):
        return f"{type(self).__name__}{stable_hashcode(self, try_cache=True).replace('-', '') if structural else str(id(self))}"

    def labeled_children(self, **kwargs):
        return [(getattr(self, f.name), f.name) for f in fields(self) if type(f.type) is type and issubclass(f.type, Symbolic)]

    def children(self, **kwargs):
        return [c for c, _ in self.labeled_children(**kwargs)]

    def vars(self) -> list[str]:
        return [v.name for v in self.preorder(lambda x: isinstance(x, Var))]

    def substitute(self, vars):
        if isinstance(self, Var) and self.name in vars:
            return vars[self.name]
        else:
            return self.map(lambda x: x.substitute(vars))

    def substitute_term(self, f):
        res = f(self)
        if res:
            return res
        else:
            return self.map(lambda x: x.substitute_term(f))

    def reconstruct(self, *cs):
        assert not self.labeled_children()
        return self

    def map(self, f):
        return self.reconstruct(*map(f, self.children())).named(self.name)

    def constant(self):
        return self.map(lambda x: None)

    def draw_node(self, **kwargs):
        print(f"{self.nodeid(**kwargs)} [label=\"{self.nodename(**kwargs)}\"];")

    def draw_edges(self, **kwargs):
        for c, label in self.labeled_children(**kwargs):
            print(f"{c.nodeid(**kwargs)} -> {self.nodeid(**kwargs)} [label=\"{label}\"];")

    def draw_children(self, **kwargs):
        for c, label in self.labeled_children(**kwargs):
            c.graphviz(**kwargs)

    def graphviz(self, structural=False, done=None, **kwargs):
        noden = self.nodeid(structural, **kwargs)
        if done is None:
            done = set()
        if noden in done:
            return
        done.add(noden)
        kwargs |= dict(done=done, structural=structural)
        self.draw_node(**kwargs)
        self.draw_edges(**kwargs)
        self.draw_children(**kwargs)

    def show_program(self, name="f", indent="    ", **kwargs):
        kwargs["indent"] = indent
        kwargs["aindent"] = indent
        kwargs["toplevel"] = True
        def extracted(x: Symbolic):
            return x.map(lambda x: x if isinstance(x, Var) else Var(f"_{subexpressions.index(x)}"))
        subexpressions = unique_by_id(self.preorder(lambda x: not isinstance(x, Var))[1:])
        subexpressions.reverse()
        free_vars = unique_by_id(self.preorder(lambda x: isinstance(x, Var)))
        arguments = [fv.show(**kwargs) for fv in free_vars]
        lines = [(x.name or f"_{i}") + " = " + extracted(x).show(**kwargs) for i, x in enumerate(subexpressions)]
        last = f"return {extracted(self).show(**kwargs)}"
        code = format_multiple([*lines, last], start=f"def {name}({format_multiple(arguments, sep=', ')}):", indent=indent, newline_threshold=0)
        return code

    def show(self, **kwargs):
        """
        Shows the expression tree code-style.
        Only works on referentially transparent DAGs.
        :param kwargs: drawing options
        :return: str
        """
        raise NotImplementedError()

    def instantiate(self, **kwargs):
        raise NotImplementedError()

    def execute(self, random: 'dict[int, AbstractBHV]' = None,
                rand2: 'dict[int, AbstractBHV]' = None,
                rand: 'dict[int, AbstractBHV]' = None,
                randomperms: 'dict[int, MemoizedPermutation]' = None,
                calculated = None, **kwargs):
        if random is None: random = {}
        if rand2 is None: rand2 = {}
        if rand is None: rand = {}
        if randomperms is None: randomperms = {}
        if calculated is None: calculated = {}
        kwargs |= dict(random=random, rand2=rand2, rand=rand, randomperms=randomperms, calculated=calculated)
        if id(self) in calculated:
            return calculated[id(self)]
        else:
            result = self.instantiate(**kwargs)
            calculated[id(self)] = result
            return result

    def optimal_sharing(self, form=None, **kwargs):
        if form is None: form = {}
        kwargs |= dict(form=form)

        sh = stable_hashcode(self, try_cache=True)
        if sh in form:
            return form[sh]
        else:
            r = self.map(lambda c: c.optimal_sharing(**kwargs))
            shr = stable_hashcode(r, try_cache=True)
            if shr in form:
                raise RuntimeError("Includes check failed")
            form[shr] = r
            return r

    def internal_size(self):
        return 1

    def size(self, counted=None, recount=False, **kwargs):
        if counted is None: counted = {}
        kwargs |= dict(counted=counted)
        if id(self) in counted:
            return counted[id(self)] if recount else 0
        else:
            t = self.internal_size()
            for c in self.children():
                t += c.size(**kwargs)
            counted[id(self)] = t
            return t

    def simplify(self, **kwargs):
        x = self
        while True:
            x_ = x.map(lambda c: c.simplify(**kwargs))
            x_r = x_.reduce(**kwargs)
            if x_r is not None:
                x_ = x_r
            if x == x_:
                return x
            else:
                x = x_

    def reduce(self, **kwargs):
        return None

    def preorder(self, p=lambda x: True):
        return [self]*p(self) + [v for c in self.children() for v in c.preorder(p)]

    def truth_assignments(self, vars):
        configs = bitconfigs(len(vars))

        return [self.execute(vars={v.name: Slice(b) for b, v in zip(config, vars)}, bhv=Slice).b
                for config in configs]


@dataclass
class List(Symbolic):
    xs: list[Symbolic]

    def show(self, **kwargs):
        return format_list((x.show(**kwargs) for x in self.xs),
                           **{k: kwargs[k] for k in ["indent", "aindent", "newline_threshold"] if k in kwargs})

    def labeled_children(self, **kwargs):
        return [(x, str(i)) for i, x in enumerate(self.xs)]

    def reconstruct(self, *cs):
        return List(cs)

    def graphviz(self, structural=False, done=None, **kwargs):
        noden = self.nodeid(structural, **kwargs)
        if done is None:
            done = set()
        if noden in done:
            return
        done.add(noden)
        kwargs |= dict(done=done, structural=structural)
        print(f"subgraph cluster_{self.nodename()} " + "{")
        print(f"label = \"{self.nodename()}\";")
        for c in self.xs:
            c.draw_node(**kwargs)
        print("}")
        for c in self.xs:
            c.draw_edges(**kwargs)
            c.draw_children(**kwargs)

    def instantiate(self, **kwargs):
        return [x.instantiate(**kwargs) for x in self.xs]

    def truth_table(self, vars):
        return [x.truth_assignments(vars) for x in self.xs]


class SymbolicPermutation(Symbolic, MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}

    @classmethod
    def random(cls) -> 'SymbolicPermutation':
        return PermRandom()

    def __mul__(self, other: 'SymbolicPermutation') -> 'SymbolicPermutation':
        return PermCompose(self, other)

    def __invert__(self) -> 'SymbolicPermutation':
        return PermInvert(self)

    def __call__(self, hv: 'SymbolicBHV') -> 'SymbolicBHV':
        return PermApply(self, hv)

@dataclass
class PermVar(SymbolicPermutation):
    name: str
    def nodename(self, **kwargs):
        return self.name

    def show(self, **kwargs):
        symbolic_var = kwargs.get("symbolic_var", False)
        return f"ParmVar(\"{self.name}\")" if symbolic_var else self.name

    def instantiate(self, **kwargs):
        permvars = kwargs.get("permvars")
        if permvars is None:
            raise RuntimeError(f"No Perm vars supplied but tried to instantiate `{self.name}`")
        elif self.name not in permvars:
            raise RuntimeError(f"Perm var `{self.name}` not in permvars ({set(permvars.keys())})")
        else:
            return permvars[self.name]
randpermid = 0
def next_perm_id():
    global randpermid
    randpermid += 1
    return randpermid
@dataclass
class PermRandom(SymbolicPermutation):
    id: int = field(default_factory=next_perm_id)

    def show(self, **kwargs):
        impl = kwargs.get("impl", "")
        random_id = kwargs.get("random_id", False)
        return f"<{impl}random {self.id}>" if random_id else impl + "random()"

    def instantiate(self, **kwargs):
        randomperms = kwargs.get("randomperms")
        if self.id in randomperms:
            return randomperms[self.id]
        else:
            r = kwargs.get("perm").random()
            randomperms[self.id] = r
            return r
@dataclass
class PermCompose(SymbolicPermutation):
    l: SymbolicPermutation
    r: SymbolicPermutation

    def reconstruct(self, l, r):
        return PermCompose(l, r)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"{self.l.show(**kwargs)} * {self.r.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs) * self.r.execute(**kwargs)
@dataclass
class PermInvert(SymbolicPermutation):
    p: SymbolicPermutation

    def reconstruct(self, p):
        return PermInvert(p)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"~{self.p.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return ~self.p.execute(**kwargs)


class SymbolicBHV(Symbolic, AbstractBHV):
    @classmethod
    def synth(cls, vs, t):
        assert 2**len(vs) == len(t)
        if vs:
            return vs[0].select(
                    cls.synth(vs[1:], t[:len(t)//2]),
                    cls.synth(vs[1:], t[len(t)//2:]))
        else:
            return cls.ONE if t[0] else cls.ZERO

    @classmethod
    def rand(cls) -> Self:
        return Rand()

    @classmethod
    def rand2(cls, power: int) -> Self:
        assert power >= 0
        return Rand2(power)

    @classmethod
    def random(cls, active: float) -> Self:
        assert 0. <= active <= 1.
        return Random(active)

    @classmethod
    def majority(cls, vs: list[Self]) -> Self:
        return Majority(vs)

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> Self:
        return Permute(permutation_id, self)

    def swap_halves(self) -> Self:
        return SwapHalves(self)

    def rehash(self) -> Self:
        return ReHash(self)

    def __xor__(self, other: Self) -> Self:
        return Xor(self, other)

    def __and__(self, other: Self) -> Self:
        return And(self, other)

    def __or__(self, other: Self) -> Self:
        return Or(self, other)

    def __invert__(self) -> Self:
        return Invert(self)

    def select(self, when1: Self, when0: Self) -> Self:
        return Select(self, when0, when1)

    def active_fraction(self) -> int:
        return ActiveFraction(self)

    def bias_rel(self, other: Self, rel: Self) -> float:
        return BiasRel(rel, self, other)

    def unrelated(self, other: Self, stdvs=6) -> bool:
        return Unrelated(self, other, stdvs)

    def expected_active_fraction(self, **kwargs):
        raise NotImplementedError()


@dataclass
class PermApply(SymbolicBHV):
    p: SymbolicPermutation
    v: SymbolicBHV

    def reconstruct(self, p, v):
        return PermApply(p, v)

    def show(self, **kwargs):
        return f"{self.p.show(**kwargs)}({self.v.show(**kwargs)})"

    def instantiate(self, **kwargs):
        return self.p.execute(**kwargs)(self.v.execute(**kwargs))

    def expected_active_fraction(self, **kwargs):
        return self.v.expected_active(**kwargs)
@dataclass
class Var(SymbolicBHV):
    name: str
    def nodename(self, **kwards):
        return self.name

    def show(self, **kwargs):
        symbolic_var = kwargs.get("symbolic_var", False)
        return f"Var(\"{self.name}\")" if symbolic_var else self.name

    def instantiate(self, **kwargs):
        vars = kwargs.get("vars")
        if vars is None:
            raise RuntimeError(f"No vars supplied but tried to instantiate `{self.name}`")
        elif self.name not in vars:
            raise RuntimeError(f"Var `{self.name}` not in vars ({set(vars.keys())})")
        else:
            return vars[self.name]

    def expected_active_fraction(self, **kwargs):
        return kwargs.get("vars").get(self.name)
@dataclass
class Zero(SymbolicBHV):
    def show(self, **kwargs):
        return kwargs.get("impl", "") + "ZERO"

    def instantiate(self, **kwargs):
        return kwargs.get("bhv").ZERO

    def expected_active_fraction(self, **kwargs):
        return 0.
@dataclass
class One(SymbolicBHV):
    def show(self, **kwargs):
        return kwargs.get("impl", "") + "ONE"

    def instantiate(self, **kwargs):
        return kwargs.get("bhv").ONE

    def expected_active_fraction(self, **kwargs):
        return 1.
SymbolicBHV.ZERO = Zero()
SymbolicBHV.ONE = One()
randid = 0
def next_id():
    global randid
    randid += 1
    return randid
@dataclass
class Rand(SymbolicBHV):
    id: int = field(default_factory=next_id)

    def show(self, **kwargs):
        impl = kwargs.get("impl", "")
        random_id = kwargs.get("random_id", False)
        return f"<{impl}rand {self.id}>" if random_id else impl + "rand()"

    def instantiate(self, **kwargs):
        rand = kwargs.get("rand")
        if self.id in rand:
            return rand[self.id]
        else:
            r = kwargs.get("bhv").rand()
            rand[self.id] = r
            return r

    def expected_active_fraction(self, **kwargs):
        return .5
@dataclass
class Rand2(SymbolicBHV):
    power: int

    def show(self, **kwargs):
        return kwargs.get("impl", "") + f"rand2({self.power})"

    def instantiate(self, **kwargs):
        rand2 = kwargs.get("rand2")
        if self.id in rand2:
            return rand2[self.id]
        else:
            r = kwargs.get("bhv").rand2(self.power)
            rand2[self.id] = r
            return r

    def expected_active_fraction(self, **kwargs):
        return 1/self.power
@dataclass
class Random(SymbolicBHV):
    frac: float

    def show(self, **kwargs):
        return kwargs.get("impl", "") + f"random({self.frac})"

    def instantiate(self, **kwargs):
        random = kwargs.get("random")
        if self.id in random:
            return random[self.id]
        else:
            r = kwargs.get("bhv").random(self.frac)
            random[self.id] = r
            return r

    def expected_active_fraction(self, **kwargs):
        return self.frac
@dataclass
class Majority(SymbolicBHV):
    vs: list[SymbolicBHV]

    def labeled_children(self, **kwargs):
        return list(zip(self.vs, map(str, range(len(self.vs)))))

    def reconstruct(self, *cs):
        return Majority(cs)

    def show(self, **kwargs):
        args = format_list((v.show(**kwargs) for v in self.vs), **{k: kwargs[k] for k in ["indent", "aindent", "newline_threshold"] if k in kwargs})
        return kwargs.get("impl", "") + f"majority({args})"

    def instantiate(self, **kwargs):
        return kwargs.get("bhv").majority([v.execute(**kwargs) for v in self.vs])

    def expected_active_fraction(self, **kwargs):
        from .poibin import PoiBin
        return 1. - PoiBin([v.expected_active_fraction(**kwargs) for v in self.vs]).cdf(len(self.vs)//2)
@dataclass
class Permute(SymbolicBHV):
    id: 'int | tuple[int, ...]'
    v: SymbolicBHV

    def reconstruct(self, v):
        return Permute(self.id, v)

    def show(self, **kwargs):
        return f"{self.v.show(**kwargs)}.permute({self.id})"

    def instantiate(self, **kwargs):
        return self.v.execute(**kwargs).permute(self.id)

    def expected_active_fraction(self, **kwargs):
        return self.v.expected_active_fraction(**kwargs)
@dataclass
class SwapHalves(SymbolicBHV):
    v: SymbolicBHV

    def reconstruct(self, v):
        return SwapHalves(v)

    def show(self, **kwargs):
        return f"{self.v.show(**kwargs)}.swap_halves()"

    def instantiate(self, **kwargs):
        return self.v.execute(**kwargs).swap_halves()

    def expected_active_fraction(self, **kwargs):
        return self.v.expected_active_fraction(**kwargs)
@dataclass
class ReHash(SymbolicBHV):
    v: SymbolicBHV

    def reconstruct(self, v):
        return ReHash(v)

    def show(self, **kwargs):
        return f"{self.v.show(**kwargs)}.rehash()"

    def instantiate(self, **kwargs):
        return self.v.execute(**kwargs).rehash()

    def expected_active_fraction(self, **kwargs):
        return .5
@dataclass
class Eq(Symbolic):
    l: SymbolicBHV
    r: SymbolicBHV

    def swap(self):
        return Eq(self.r, self.l)

    def reconstruct(self, l, r):
        return Eq(l, r)

    def show(self, toplevel=True, **kwargs):
        brackets = not toplevel
        kwargs["toplevel"] = False
        return "("*brackets + f"{self.l.show(**kwargs)} == {self.r.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs) == self.r.execute(**kwargs)
@dataclass
class Xor(SymbolicBHV):
    l: SymbolicBHV
    r: SymbolicBHV

    def reconstruct(self, l, r):
        return Xor(l, r)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"{self.l.show(**kwargs)} ^ {self.r.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs) ^ self.r.execute(**kwargs)

    def reduce(self, **kwargs):
        if self.l == self.ONE:
            return ~self.r
        elif self.r == self.ONE:
            return ~self.l
        elif self.l == self.ZERO:
            return self.r
        elif self.r == self.ZERO:
            return self.l
        elif self.l == self.r:
            return self.ZERO
        elif self.l == ~self.r or ~self.l == self.r:
            return self.ONE
        elif isinstance(self.l, Invert) and isinstance(self.r, Invert):
            return Xor(self.l.v, self.r.v)
        elif isinstance(self.l, And) and isinstance(self.r, And):
            if self.l.l == self.r.l: return And(self.l.l, Xor(self.l.r, self.r.r))
            elif self.l.l == self.r.r: return And(self.l.l, Xor(self.l.r, self.r.l))
            elif self.l.r == self.r.l: return And(self.l.r, Xor(self.l.l, self.r.r))
            elif self.l.r == self.r.r: return And(self.l.r, Xor(self.l.l, self.r.l))

    def expected_active_fraction(self, **kwargs):
        afl = self.l.expected_active_fraction(**kwargs)
        afr = self.r.expected_active_fraction(**kwargs)
        return afl*(1. - afr) + (1. - afl)*afr
@dataclass
class And(SymbolicBHV):
    l: SymbolicBHV
    r: SymbolicBHV

    def reconstruct(self, l, r):
        return And(l, r)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"{self.l.show(**kwargs)} & {self.r.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs) & self.r.execute(**kwargs)

    def reduce(self, **kwargs):
        if self.l == self.ZERO or self.r == self.ZERO:
            return self.ZERO
        elif self.l == self.ONE:
            return self.r
        elif self.r == self.ONE:
            return self.l
        elif isinstance(self.l, Invert) and self.l.v == self.r:
            return self.ZERO
        elif isinstance(self.r, Invert) and self.r.v == self.l:
            return self.ZERO
        elif isinstance(self.l, Invert) and isinstance(self.r, Invert):
            return Invert(Or(self.l.v, self.r.v))

    def expected_active_fraction(self, **kwargs):
        afl = self.l.expected_active_fraction(**kwargs)
        afr = self.r.expected_active_fraction(**kwargs)
        return afl*afr
@dataclass
class Or(SymbolicBHV):
    l: SymbolicBHV
    r: SymbolicBHV

    def reconstruct(self, l, r):
        return Or(l, r)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"{self.l.show(**kwargs)} | {self.r.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs) | self.r.execute(**kwargs)

    def reduce(self, **kwargs):
        if self.l == self.ONE or self.r == self.ONE:
            return self.ONE
        elif self.l == self.ZERO:
            return self.r
        elif self.r == self.ZERO:
            return self.l
        elif isinstance(self.l, Invert) and self.l.v == self.r:
            return self.ONE
        elif isinstance(self.r, Invert) and self.r.v == self.l:
            return self.ONE
        elif isinstance(self.l, Invert) and isinstance(self.r, Invert):
            return Invert(And(self.l.v, self.r.v))

    def expected_active_fraction(self, **kwargs):
        afl = self.l.expected_active_fraction(**kwargs)
        afr = self.r.expected_active_fraction(**kwargs)
        return 1. - ((1. - afl)*(1. - afr))
@dataclass
class Invert(SymbolicBHV):
    v: SymbolicBHV

    def reconstruct(self, v):
        return Invert(v)

    def show(self, **kwargs):
        brackets = not kwargs.get("toplevel", False)
        return "("*brackets + f"~{self.v.show(**kwargs)}" + ")"*brackets

    def instantiate(self, **kwargs):
        return ~self.v.execute(**kwargs)

    def reduce(self, **kwargs):
        if self.v == self.ONE:
            return self.ZERO
        elif self.v == self.ZERO:
            return self.ONE
        elif isinstance(self.v, Invert):
            return self.v.v

    def expected_active_fraction(self, **kwargs):
        return 1. - self.v.expected_active_fraction(**kwargs)
@dataclass
class Select(SymbolicBHV):
    cond: SymbolicBHV
    when1: SymbolicBHV
    when0: SymbolicBHV

    def reconstruct(self, c, w1, w0):
        return Select(c, w1, w0)

    def nodename(self, compact_select=False, **kwargs):
        return f"ON {self.cond.nodename()}" if compact_select else super().nodename(**kwargs)

    def labeled_children(self, compact_select=False, **kwargs):
        return [(self.when1, "1"), (self.when0, "0")] if compact_select else super().labeled_children(**kwargs)

    def show(self, **kwargs):
        return f"{self.cond.show(**kwargs)}.select({self.when1.show(**kwargs)}, {self.when0.show(**kwargs)})"

    def instantiate(self, **kwargs):
        return self.cond.execute(**kwargs).select(self.when1.execute(**kwargs), self.when0.execute(**kwargs))

    def internal_size(self):
        return 3

    def reduce(self, **kwargs):
        expand_select_xor = kwargs.get("expand_select_xor", False)
        expand_select_and_or = kwargs.get("expand_select_and_or", False)

        if self.when1 == self.ONE and self.when0 == self.ZERO:
            return self.cond
        elif self.when1 == self.ZERO and self.when0 == self.ONE:
            return ~self.cond
        elif self.when0 == self.when1:
            return self.when0
        elif self.when1 == self.ONE:
            return self.cond | self.when0
        elif self.when1 == self.ZERO:
            return ~self.cond & self.when0
        elif self.when0 == self.ONE:
            return ~self.cond | self.when1
        elif self.when0 == self.ZERO:
            return self.cond & self.when1
        else:
            if self.when1 == ~self.when0:
                return self.cond ^ self.when0
            elif self.when0 == ~self.when1:
                return self.cond ^ self.when0
            elif isinstance(self.when0, Invert) and isinstance(self.when1, Invert):
                return ~Select(self.cond, self.when1.v, self.when0.v)
            else:
                if expand_select_xor:
                    return self.when0 ^ (self.cond & (self.when0 ^ self.when1))
                elif expand_select_and_or:
                    return (self.cond & self.when1) | (~self.cond & self.when0)

    def expected_active_fraction(self, **kwargs):
        afc = self.cond.expected_active_fraction(**kwargs)
        af1 = self.when1.expected_active_fraction(**kwargs)
        af0 = self.when0.expected_active_fraction(**kwargs)
        return afc*af1 + (1. - afc)*af0
@dataclass
class ActiveFraction(Symbolic):
    v: SymbolicBHV

    def reconstruct(self, v):
        return ActiveFraction(v)

    def show(self, **kwargs):
        return f"{self.v.show(**kwargs)}.active_fraction()"

    def instantiate(self, **kwargs):
        return self.v.execute(**kwargs).active_fraction()
@dataclass
class BiasRel(Symbolic):
    rel: SymbolicBHV
    l: SymbolicBHV
    r: SymbolicBHV

    def reconstruct(self, rel, l, r):
        return BiasRel(rel, l, r)

    def show(self, **kwargs):
        return f"{self.l.show(**kwargs)}.bias_rel({self.r.show(**kwargs)}, {self.rel.show(**kwargs)})"

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs).bias_rel(self.r.execute(**kwargs), self.rel.execute(**kwargs))
@dataclass
class Related(Symbolic):
    l: SymbolicBHV
    r: SymbolicBHV
    stdvs: float

    def reconstruct(self, l, r):
        return Related(l, r, self.stdvs)

    def show(self, **kwargs):
        return f"{self.l.show(**kwargs)}.related({self.r.show(**kwargs)}, {self.stdvs})"

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs).related(self.r.execute(**kwargs), self.stdvs)
@dataclass
class Unrelated(Symbolic):
    l: SymbolicBHV
    r: SymbolicBHV
    stdvs: float

    def reconstruct(self, l, r):
        return Unrelated(l, r, self.stdvs)

    def show(self, **kwargs):
        return f"{self.l.show(**kwargs)}.unrelated({self.r.show(**kwargs)}, {self.stdvs})"

    def instantiate(self, **kwargs):
        return self.l.execute(**kwargs).unrelated(self.r.execute(**kwargs), self.stdvs)



from .abstract import *
import torch


# GPT-4 wrote this code, do not use in production
def pack_bool_to_long(bool_tensor):
    assert bool_tensor.dtype == torch.bool, "Input tensor must be of dtype torch.bool"
    assert bool_tensor.numel() % 64 == 0, "Input tensor must have a number of elements divisible by 64"

    bool_tensor = bool_tensor.view(-1, 64)
    packed_tensor = torch.zeros(bool_tensor.shape[0], dtype=torch.int64)

    for i in range(64):
        packed_tensor |= (bool_tensor[:, i].to(torch.int64) << i)

    return packed_tensor


# GPT-4 wrote this code, do not use in production
def unpack_long_to_bool(packed_tensor):
    assert packed_tensor.dtype == torch.int64, "Input tensor must be of dtype torch.int64"

    bool_tensor = torch.zeros(packed_tensor.numel() * 64, dtype=torch.bool)

    for i in range(64):
        bool_tensor[i::64] = ((packed_tensor >> i) & 1).bool()

    return bool_tensor


class TorchBoolPermutation(MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}

    def __init__(self, tensor: torch.IntTensor):
        self.data: torch.BoolTensor = tensor

    @classmethod
    def random(cls) -> 'TorchBoolPermutation':
        return TorchBoolPermutation(torch.randperm(DIMENSION))

    def __mul__(self, other: 'TorchBoolPermutation') -> 'TorchBoolPermutation':
        return TorchBoolPermutation(self.data[other.data])

    def __invert__(self) -> 'TorchBoolPermutation':
        inv_permutation = torch.empty_like(self.data)
        inv_permutation[self.data] = torch.arange(DIMENSION)
        return TorchBoolPermutation(inv_permutation)

    def __call__(self, hv: 'TorchBoolBHV') -> 'TorchBoolBHV':
        return hv.permute_bits(self)

TorchBoolPermutation.IDENTITY = TorchBoolPermutation(torch.arange(DIMENSION))


class TorchBoolBHV(AbstractBHV):
    def __init__(self, tensor: torch.BoolTensor):
        self.data: torch.BoolTensor = tensor

    @classmethod
    def rand(cls) -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.empty(DIMENSION, dtype=torch.bool).random_())

    @classmethod
    def random(cls, active: float) -> 'TorchBoolBHV':
        assert 0. <= active <= 1.
        return TorchBoolBHV(torch.empty(DIMENSION, dtype=torch.bool).bernoulli_(active))

    def select(self, when1: 'TorchBoolBHV', when0: 'TorchBoolBHV') -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.where(self.data, when1.data, when0.data))

    @classmethod
    def majority(cls, vs: list['TorchBoolBHV']) -> 'TorchBoolBHV':
        data = [v.data for v in vs]
        extra = [cls.rand().data] if len(vs) % 2 == 0 else []

        tensor = torch.stack(data + extra)
        counts = tensor.sum(dim=-2, dtype=torch.uint8 if len(vs) < 256 else torch.int32)

        threshold = (len(vs) + len(extra))//2

        return TorchBoolBHV(torch.greater(counts,  threshold).to(torch.bool))

    def roll_bits(self, n: int) -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.roll(self.data, n))

    def permute_bits(self, permutation: TorchBoolPermutation) -> 'TorchBoolBHV':
        return TorchBoolBHV(self.data[permutation.data])

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> 'TorchBoolBHV':
        return self.permute_bits(TorchBoolPermutation.get(permutation_id))

    def swap_halves(self) -> 'TorchBoolBHV':
        return self.roll_bits(DIMENSION//2)

    def rehash(self) -> 'TorchBoolBHV':
        # torch to bytes is broken hence custom function https://github.com/pytorch/pytorch/issues/33041
        offsets = [(H & self).active() for H in self._HS]
        res = self._HASH
        for o in offsets:
            res = res ^ self.roll_bits(o)
        return res

    def __eq__(self, other: 'TorchBoolBHV') -> bool:
        return torch.equal(self.data, other.data)

    def __xor__(self, other: 'TorchBoolBHV') -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.bitwise_xor(self.data, other.data))

    def __and__(self, other: 'TorchBoolBHV') -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.bitwise_and(self.data, other.data))

    def __or__(self, other: 'TorchBoolBHV') -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.bitwise_or(self.data, other.data))

    def __invert__(self) -> 'TorchBoolBHV':
        return TorchBoolBHV(torch.bitwise_not(self.data))

    def active(self) -> int:
        return torch.sum(self.data).item()

    def pack(self) -> 'TorchPackedBHV':
        return TorchPackedBHV(pack_bool_to_long(self.data))

TorchBoolBHV.ZERO = TorchBoolBHV(torch.zeros(DIMENSION, dtype=torch.bool))
TorchBoolBHV.ONE = TorchBoolBHV(torch.ones(DIMENSION, dtype=torch.bool))
TorchBoolBHV._HASH = TorchBoolBHV.rand()
TorchBoolBHV._HS = TorchBoolBHV.nrand2(5, power=2)
TorchBoolBHV._FEISTAL_SUBKEYS = TorchBoolBHV.nrand2(TorchBoolBHV._FEISTAL_ROUNDS, 4)
_halfb = torch.zeros(DIMENSION, dtype=torch.bool)
_halfb[:DIMENSION//2] = 1
TorchBoolBHV.HALF = TorchBoolBHV(_halfb)


class TorchWordPermutation(MemoizedPermutation):
    _permutations: 'dict[int | tuple[int, ...], Self]' = {}

    def __init__(self, tensor: torch.IntTensor):
        self.data: torch.BoolTensor = tensor

    @classmethod
    def random(cls) -> 'TorchWordPermutation':
        return TorchWordPermutation(torch.randperm(DIMENSION//64))

    def __mul__(self, other: 'TorchWordPermutation') -> 'TorchWordPermutation':
        return TorchWordPermutation(self.data[other.data])

    def __invert__(self) -> 'TorchWordPermutation':
        inv_permutation = torch.empty_like(self.data)
        inv_permutation[self.data] = torch.arange(DIMENSION//64)
        return TorchWordPermutation(inv_permutation)

    def __call__(self, hv: 'TorchPackedBHV') -> 'TorchPackedBHV':
        return hv.permute_words(self)

TorchWordPermutation.IDENTITY = TorchWordPermutation(torch.arange(DIMENSION//64))


class TorchPackedBHV(AbstractBHV):
    def __init__(self, tensor: torch.LongTensor):
        assert DIMENSION % 64 == 0
        self.data: torch.LongTensor = tensor

    @classmethod
    def rand(cls) -> Self:
        return TorchPackedBHV(torch.randint(-9223372036854775808, 9223372036854775807, size=(DIMENSION//64,), dtype=torch.long))

    @classmethod
    def random(cls, active) -> Self:
        assert 0. <= active <= 1.
        return TorchBoolBHV.random(active).pack()

    def roll_words(self, n: int) -> 'TorchPackedBHV':
        return TorchPackedBHV(torch.roll(self.data, n))

    def permute_words(self, permutation: TorchWordPermutation) -> 'TorchPackedBHV':
        return TorchPackedBHV(self.data[permutation.data])

    def permute(self, permutation_id: 'int | tuple[int, ...]') -> 'TorchPackedBHV':
        return self.permute_words(TorchWordPermutation.get(permutation_id))

    def swap_halves(self) -> 'TorchPackedBHV':
        return self.roll_words(DIMENSION//128)

    def rehash(self) -> 'TorchPackedBHV':
        return self.unpack().rehash().pack()

    def __eq__(self, other: 'TorchBoolBHV') -> bool:
        return torch.equal(self.data, other.data)

    def __xor__(self, other: 'TorchPackedBHV') -> 'TorchPackedBHV':
        return TorchPackedBHV(torch.bitwise_xor(self.data, other.data))

    def __and__(self, other: 'TorchPackedBHV') -> 'TorchPackedBHV':
        return TorchPackedBHV(torch.bitwise_and(self.data, other.data))

    def __or__(self, other: 'TorchPackedBHV') -> 'TorchPackedBHV':
        return TorchPackedBHV(torch.bitwise_or(self.data, other.data))

    def __invert__(self) -> 'TorchPackedBHV':
        return TorchPackedBHV(torch.bitwise_not(self.data))

    def active(self) -> int:
        # currently no efficient implementation available for this https://github.com/pytorch/pytorch/issues/36380
        x = self.data
        x = x - ((x >> 1) & 0x5555555555555555)
        x = (x & 0x3333333333333333) + ((x >> 2) & 0x3333333333333333)
        x = (x + (x >> 4)) & 0x0f0f0f0f0f0f0f0f
        x = (x * 0x0101010101010101) >> 56
        return torch.sum(x).item()

    def unpack(self) -> 'TorchBoolBHV':
        return TorchBoolBHV(unpack_long_to_bool(self.data))

TorchPackedBHV.ZERO = TorchPackedBHV(torch.zeros(DIMENSION//64, dtype=torch.long))
TorchPackedBHV.ONE = TorchPackedBHV(torch.full((DIMENSION//64,), fill_value=-1, dtype=torch.long))  # -1 is all ones in torch's encoding
TorchPackedBHV._FEISTAL_SUBKEYS = TorchPackedBHV.nrand2(TorchPackedBHV._FEISTAL_ROUNDS, 4)
_half64 = torch.zeros(DIMENSION//64, dtype=torch.long)
_half64[:DIMENSION//128] = -1
TorchPackedBHV.HALF = TorchPackedBHV(_half64)
